/*
 * test.cu
 *
 *  Created on: February 27, 2011
 *      Author: Alex Krizhevsky (akrizhevsky@gmail.com)
 */

#include "../include/localCPU.h"

#include "../include/test.cuh"

static uint timer;

void init_tests(int boardNum) {
    cudaSetDevice(boardNum > -1 ? boardNum : cutGetMaxGflopsDeviceId());
    cublasInit();
    NVMatrix::initRandom(7);
    cutilCheckError(cutCreateTimer( &timer));
}

void test_filter_acts() {
    printf("===============================\n");
    printf("test_filter_acts\n");
    printf("===============================\n");

    FILTER_OUTPUT_ORDER targetsOrder = FILTER_MODULE_IMAGE;
    int colors = 2;
    int filterSize = 25, imgSize = 96;
    int numFilters = 32;
    int moduleStride = 5;
    // Compute minimal number of filter applications to cover entire image
    int numModulesX = 1 + int(ceil((imgSize - filterSize) / float(moduleStride)));
    int numModules = numModulesX * numModulesX;
    int numImages = 128;
    // Pad the images with this many zeros on the top and on the left.
    int paddingStart = -((numModulesX - 1) * moduleStride + filterSize - imgSize) / 2;
    int imgPixels = imgSize * imgSize;
    int filterPixels = filterSize * filterSize;
    printf("Filters per module: %d\n",  numFilters);
    printf("Modules: %dx%d\n",  numModulesX, numModulesX);
    printf("Images: %d %dx%d\n",  numImages, imgSize, imgSize);
    printf("Filter size: %d\n", filterSize);
    printf("Padding start: %d\n", paddingStart);
    printf("Module stride: %d\n", moduleStride);
    printf("Colors: %d\n", colors);

    Matrix images;
    Matrix filters(colors*numModules * filterPixels, numFilters);
    Matrix targets(numFilters * numModules, numImages);

    filters.randomizeUniform();
    targets.apply(Matrix::ZERO);

    filters.addScalar(-0.5);

    NVMatrix nvFilters(filters, true);
    NVMatrix nvImagesSrc(2*imgPixels*colors, numImages * 4, false);
    /*
     * This is just for testing that everything works fine when nvImages is a non-contiguous matrix.
     */
    NVMatrix& nvImages = nvImagesSrc.slice(0, colors*imgPixels, numImages, numImages * 2);
    NVMatrix nvTargets(targets, true);

    nvImagesSrc.randomizeUniform();
    nvImages.copyToHost(images, true);

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    cpuFilterActs(images.getData(), filters.getData(), targets.getData(), numImages, numFilters,
            imgSize, filterSize, paddingStart, moduleStride, numModulesX, colors, targetsOrder);

    cutilCheckError(cutStopTimer(timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 5, 0, 5);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    computeFilterActs(nvImages, nvFilters, nvTargets, numModulesX, paddingStart, moduleStride, colors, targetsOrder);

    cudaThreadSynchronize();
    cutilCheckError(cutStopTimer(timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 5, 0, 5);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_weight_acts() {
    printf("===============================\n");
    printf("test_weight_acts\n");
    printf("===============================\n");

    FILTER_OUTPUT_ORDER hidActsOrder = FILTER_MODULE_IMAGE;
    int colors = 2;
    int filterSize = 11, imgSize = 96;
    int numFilters = 32;
    int moduleStride = 5;
    // Compute minimal number of filter applications to cover entire image
    int numModulesX = 1 + int(ceil((imgSize - filterSize) / float(moduleStride)));
    int numModules = numModulesX * numModulesX;
    int numImages = 128;
    // Pad the images with this many zeros on the top and on the left.
    int paddingStart = -((numModulesX - 1) * moduleStride + filterSize - imgSize) / 2;
    int imgPixels = imgSize * imgSize;
    int filterPixels = filterSize * filterSize;
    printf("Filters per module: %d\n",  numFilters);
    printf("Modules: %dx%d\n",  numModulesX, numModulesX);
    printf("Images: %d %dx%d\n",  numImages, imgSize, imgSize);
    printf("Filter size: %d\n", filterSize);
    printf("Padding start: %d\n", paddingStart);
    printf("Module stride: %d\n", moduleStride);
    printf("Colors: %d\n", colors);

    Matrix images;
    Matrix hidActs(numModules * numFilters, numImages);
    Matrix targets(colors*numModules * filterPixels, numFilters);

    hidActs.randomizeUniform();
    targets.apply(Matrix::ZERO);

    NVMatrix nvHidActs(hidActs, true);
    NVMatrix nvImagesSrc(2*imgPixels*colors, numImages * 4, false);
    /*
     * This is just for testing that everything works fine when nvImages is a non-contiguous matrix.
     */
    NVMatrix& nvImages = nvImagesSrc.slice(0, colors*imgPixels, numImages, numImages * 2);
    NVMatrix nvTargets(targets, true); 

    nvImagesSrc.randomizeUniform();
    nvImages.copyToHost(images, true);

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    cpuWeightActs(images.getData(), hidActs.getData(), targets.getData(), numImages, numFilters,
                        imgSize, filterSize, numModulesX, paddingStart, moduleStride, colors, hidActsOrder);

    cutilCheckError(cutStopTimer(timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 5, 0, 5);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    computeWeightActs(nvImages, nvHidActs, nvTargets, numModulesX, filterSize, paddingStart, moduleStride, colors, hidActsOrder);

    cudaThreadSynchronize();
    cutilCheckError(cutStopTimer(timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 5, 0, 5);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_img_acts() {
    printf("===============================\n");
    printf("test_img_acts\n");
    printf("===============================\n");

    FILTER_OUTPUT_ORDER hidActsOrder = FILTER_MODULE_IMAGE;
    int colors = 4;
    int filterSize = 35, imgSize = 96;
    int numFilters = 32;
    int moduleStride = 5;
    // Compute minimal number of filter applications to cover entire image
    int numModulesX = 1 + int(ceil((imgSize - filterSize) / float(moduleStride)));
    int numModules = numModulesX * numModulesX;
    int numImages = 128;
    // Pad the images with this many zeros on the top and on the left.
    int paddingStart = -((numModulesX - 1) * moduleStride + filterSize - imgSize) / 2;
    int imgPixels = imgSize * imgSize;
    int filterPixels = filterSize * filterSize;
    printf("Filters per module: %d\n",  numFilters);
    printf("Modules: %dx%d\n",  numModulesX, numModulesX);
    printf("Images: %d %dx%d\n",  numImages, imgSize, imgSize);
    printf("Filter size: %d\n", filterSize);
    printf("Padding start: %d\n", paddingStart);
    printf("Module stride: %d\n", moduleStride);
    printf("Colors: %d\n", colors);

    Matrix hidActs(numModules  * numFilters, numImages);
    Matrix weights(numModules * colors*filterPixels, numFilters);
    Matrix targets(colors*imgPixels, numImages);

    weights.randomizeUniform();
    hidActs.randomizeUniform();
    targets.apply(Matrix::ZERO);

    NVMatrix nvWeights(weights, true);
    NVMatrix nvHidActs(hidActs, true);
    NVMatrix nvTargets(targets, true);

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    cpuImgActs(hidActs.getData(), weights.getData(),targets.getData(),numModulesX,numImages,
            numFilters,filterSize, imgSize, paddingStart, moduleStride, colors, hidActsOrder);

    cutilCheckError(cutStopTimer(timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 5, 0, 5);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError(cutResetTimer(timer));
    cutilCheckError(cutStartTimer(timer));

    computeImgActs(nvHidActs, nvWeights, nvTargets, imgSize, paddingStart, moduleStride, colors,hidActsOrder);

    cudaThreadSynchronize();
    cutilCheckError(cutStopTimer(timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 5, 0, 5);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

int main(int argc, char** argv) {
    int boardNum = get_board_lock();
    if (boardNum == GPU_LOCK_NO_BOARD) {
        printf("No free GPU boards!\n");
        exit(EXIT_FAILURE);
    } else if(boardNum == GPU_LOCK_NO_SCRIPT) {
        printf("Running on default board.\n");
    } else {
        printf("Running on board %d\n", boardNum);
    }
    init_tests(boardNum);

    test_filter_acts();
    test_img_acts();
    test_weight_acts();

    return 0;
}